/*
 *
 * Copyright (C) 2019-2021, Broadband Forum
 * Copyright (C) 2016-2021  CommScope, Inc
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * \file dm_inst_vector.c
 *
 * Implements a data structure containing a list of dm_inst structures
 * This is basically a list of all object instances instantiated in the data model
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common_defs.h"
#include "data_model.h"
#include "int_vector.h"
#include "dm_inst_vector.h"


//--------------------------------------------------------------------
// Top-level multi-instance node being refreshed
// This is used to ensure that within the refresh instances callback,
// the caller of USP_RefreshInstance() is only refreshing the instances that it has ben asked to refresh
// Outside of the refresh instances callback, this variable is set to NULL
static dm_node_t *refresh_instances_top_node = NULL;

//--------------------------------------------------------------------
// Vector, used to hold the new set of instances for the refresh_instances_top_node, within the refresh instances callback
static dm_instances_vector_t refreshed_instances_vector = { 0 };

//--------------------------------------------------------------------
// Forward declarations. Note these are not static, because we need them in the symbol table for USP_LOG_Callstack() to show them
void AddObjectInstanceIfPermitted(dm_instances_t *inst, str_vector_t *sv, combined_role_t *combined_role);
int RefreshInstVector(dm_node_t *node, bool notify_subscriptions);
int RefreshInstVectorEntry(char *path);
bool IsExistInInstVector(dm_instances_t *match, dm_instances_vector_t *div);
void AddToInstVector(dm_instances_t *inst, dm_instances_vector_t *div);

/*********************************************************************//**
**
** DM_INST_VECTOR_Init
**
** Initialises a dm_inst vector
**
** \param   div - pointer to dm_instances vector structure to initialize
**
** \return  None
**
**************************************************************************/
void DM_INST_VECTOR_Init(dm_instances_vector_t *div)
{
    div->vector = NULL;
    div->num_entries = 0;
}

/*********************************************************************//**
**
** DM_INST_VECTOR_Destroy
**
** Frees up all memory used by the specified dm_instances_vector structure
**
** \param   div - pointer to dm_instances vector structure
**
** \return  None
**
**************************************************************************/
void DM_INST_VECTOR_Destroy(dm_instances_vector_t *div)
{
    if (div->vector != NULL)
    {
        USP_FREE(div->vector);
    }

    div->vector = NULL;
    div->num_entries = 0;
}

/*********************************************************************//**
**
** DM_INST_VECTOR_Add
**
** Adds the specified instance numbers and associated nodes to the dm_instances_vector vector
** NOTE: The instance is not added again, if it already exists
**
** \param   inst - pointer to instance structure to add to the dm_instances_vector vector
**                 contained within this structure is the top level multi-instance node which holds the dm_instances_vector
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
int DM_INST_VECTOR_Add(dm_instances_t *inst)
{
    int i;
    dm_instances_t *oi;
    dm_node_t *top_node;
    dm_instances_vector_t *div;

    // Exit if there are no object instances to add
    // This is the case if this function is called for a parameter which does not have any object instances in it's path
    if (inst->order == 0)
    {
        return USP_ERR_OK;
    }

    // Determine which top level multi-instance node's DM instances array to add to
    USP_ASSERT(inst->order > 0);
    top_node = inst->nodes[0];
    USP_ASSERT(top_node != NULL);
    USP_ASSERT(top_node->type == kDMNodeType_Object_MultiInstance);
    div = &top_node->registered.object_info.inst_vector;

    // See if this instance already exists
    for (i=0; i < div->num_entries; i++)
    {
        // If this instance of the object already exists then exit, nothing more to do
        oi = &div->vector[i];
        if (memcmp(oi, inst, sizeof(dm_instances_t)) == 0)
        {
            return USP_ERR_OK;
        }
    }

    // Add the instance to the correct top-level node instance vector
    AddToInstVector(inst, div);

    return USP_ERR_OK;
}


/*********************************************************************//**
**
** DM_INST_VECTOR_Remove
**
** Deletes the specified instance numbers and associated nodes from the dm_instances_vector vector
** NOTE: This function deletes the instance number tree starting at the specified instance
** NOTE: The instance is not removed again, if it has already been removed
**
** \param   inst - pointer to instance structure to delete from the dm_instances_vector vector
**                 contained within this structure is the top level multi-instance node which holds the dm_instances_vector
**
** \return  None
**
**************************************************************************/
void DM_INST_VECTOR_Remove(dm_instances_t *inst)
{
    int i;
    int j;
    int order;
    dm_instances_t *oi;
    dm_node_t *top_node;
    dm_instances_vector_t *div;

    // Exit if there is no instance to remove
    if (inst->order == 0)
    {
        return;
    }

    // Determine which top level multi-instance node's DM instances array to remove from
    USP_ASSERT(inst->order > 0);
    top_node = inst->nodes[0];
    USP_ASSERT(top_node != NULL);
    USP_ASSERT(top_node->type == kDMNodeType_Object_MultiInstance);
    div = &top_node->registered.object_info.inst_vector;

    // Find this instance and all child nested instances and delete them
    j = 0;
    order = inst->order;
    for (i=0; i < div->num_entries; i++)
    {
        oi = &div->vector[i];
        if ((oi->order >= order) &&
            (memcmp(oi->nodes, inst->nodes, order*sizeof(dm_node_t *)) == 0) &&
            (memcmp(oi->instances, inst->instances, order*sizeof(int)) == 0))
        {
            // Delete this node. Nothing to do in this iteration of the loop - this value will be overwritten by further
        }
        else
        {
            // Copy down later entries in the array, over ones which have been removed
            if (j < i)
            {
                memcpy(&div->vector[j], oi, sizeof(dm_instances_t));
            }

            j++;
        }
    }

    // NOTE: Don't bother reallocating the memory for the array (it could now be smaller).
    // It will be resized next time an instance is added.
    div->num_entries = j;
}

/*********************************************************************//**
**
** DM_INST_VECTOR_IsExist
**
** Determines whether the specified object instance exists in the data model
**
** \param   match - pointer to instances structure describing the instances to match against
**                 contained within this structure is the top level multi-instance node which holds the dm_instances_vector
** \param   exists - pointer to boolean in which to return whether the object exists or not
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
int DM_INST_VECTOR_IsExist(dm_instances_t *match, bool *exists)
{
    dm_node_t *top_node;
    dm_instances_vector_t *div;
    int err;

    // Exit if the object is a single instance object or an unqualified multi-instance object - these always exist
    if (match->order == 0)
    {
        *exists = true;
        return USP_ERR_OK;
    }

    // Determine which top level multi-instance node's DM instances array to search in
    USP_ASSERT(match->order > 0);
    top_node = match->nodes[0];
    USP_ASSERT(top_node != NULL);
    USP_ASSERT(top_node->type == kDMNodeType_Object_MultiInstance);
    div = &top_node->registered.object_info.inst_vector;

    // Exit if unable to refresh the cache of instances for this object (if necessary)
    err = RefreshInstVector(top_node, true);
    if (err != USP_ERR_OK)
    {
        return err;
    }

    *exists = IsExistInInstVector(match, div);

    return USP_ERR_OK;
}

/*********************************************************************//**
**
** DM_INST_VECTOR_GetNextInstance
**
** Gets the next numbered instance for the specified object (given it's parent instance numbers)
** This function is called when allocating a new instance number for an object that is being added
**
** \param   node - pointer to object in data model
** \param   inst - pointer to instance structure specifying the object's parents and their instance numbers
**                 contained within this structure is the top level multi-instance node which holds the dm_instances_vector
** \param   next_instance - pointer to variable in which to return the next instance number
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
int DM_INST_VECTOR_GetNextInstance(dm_node_t *node, dm_instances_t *inst, int *next_instance)
{
    int i;
    int order;
    int instance;
    int highest_instance=0;       // highest instance number encountered so far
    dm_instances_t *oi;
    dm_node_t *top_node;
    dm_instances_vector_t *div;

    order = inst->order;            // NOTE: This may be 0 for a top level multi-instance node
    USP_ASSERT(order < MAX_DM_INSTANCE_ORDER);
    inst->nodes[order] = node;

    // Determine which top level multi-instance node's DM instances array to iterate over
    top_node = inst->nodes[0];
    USP_ASSERT(top_node != NULL);
    USP_ASSERT(top_node->type == kDMNodeType_Object_MultiInstance);
    div = &top_node->registered.object_info.inst_vector;

    // Iterate over the table of instance numbers, determining the highest instance number for the specified object
    for (i=0; i < div->num_entries; i++)
    {
        oi = &div->vector[i];
        if ((oi->order == order+1) &&
            (memcmp(oi->nodes, inst->nodes, (order+1)*sizeof(dm_node_t *)) == 0) &&
            (memcmp(oi->instances, inst->instances, order*sizeof(int)) == 0))
        {
            instance = oi->instances[order];
            if (instance > highest_instance)
            {
                highest_instance = instance;
            }
        }
    }

    *next_instance = highest_instance+1;
    inst->nodes[order] = NULL;          // Undo the changes made by this function to the inst array

    return USP_ERR_OK;
}

/*********************************************************************//**
**
** DM_INST_VECTOR_GetNumInstances
**
** Gets the number of instances of the specified object (given it's parent instance numbers)
**
** \param   node - pointer to object in data model
** \param   inst - pointer to instance structure specifying the object's parents and their instance numbers
** \param   num_instances - pointer to variable in which to return the number of instances of the specified object
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
int DM_INST_VECTOR_GetNumInstances(dm_node_t *node, dm_instances_t *inst, int *num_instances)
{
    int i;
    int order;
    int count;
    int err;
    dm_instances_t *oi;
    dm_node_t *top_node;
    dm_instances_vector_t *div;

    order = inst->order;           // NOTE: This may be 0 for a top level multi-instance node
    USP_ASSERT(order < MAX_DM_INSTANCE_ORDER);
    inst->nodes[order] = node;

    // Determine which top level multi-instance node's DM instances array to iterate over
    top_node = inst->nodes[0];
    USP_ASSERT(top_node != NULL);
    USP_ASSERT(top_node->type == kDMNodeType_Object_MultiInstance);
    div = &top_node->registered.object_info.inst_vector;

    // Exit if unable to refresh the cache of instances for this object (if necessary)
    err = RefreshInstVector(top_node, true);
    if (err != USP_ERR_OK)
    {
        return err;
    }

    // Iterate over the table of instance numbers, counting the number of instances which match the object and its parent instance numbers
    count = 0;
    for (i=0; i < div->num_entries; i++)
    {
        oi = &div->vector[i];
        if ((oi->order == order+1) &&
            (memcmp(oi->nodes, inst->nodes, (order+1)*sizeof(dm_node_t *)) == 0) &&
            (memcmp(oi->instances, inst->instances, order*sizeof(int)) == 0))
        {
            count++;
        }
    }

    inst->nodes[order] = NULL;          // Undo the changes made by this function to the inst array

    *num_instances = count;

    return USP_ERR_OK;
}

/*********************************************************************//**
**
** DM_INST_VECTOR_GetInstances
**
** Gets a vector of the instance numbers for the specified object (given it's parent instance numbers)
**
** \param   node - pointer to object in data model
** \param   inst - pointer to instance structure specifying the object's parents and their instance numbers
** \param   iv - pointer to structure which will be populated with instance numbers by this function
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
int DM_INST_VECTOR_GetInstances(dm_node_t *node, dm_instances_t *inst, int_vector_t *iv)
{
    int i;
    int order;
    int instance;
    dm_instances_t *oi;
    int index;
    int err;
    dm_node_t *top_node;
    dm_instances_vector_t *div;

    order = inst->order;          // NOTE: This may be 0 for a top level multi-instance node
    USP_ASSERT(order < MAX_DM_INSTANCE_ORDER);
    inst->nodes[order] = node;
    INT_VECTOR_Init(iv);

    // Determine which top level multi-instance node's DM instances array to iterate over
    top_node = inst->nodes[0];
    USP_ASSERT(top_node != NULL);
    USP_ASSERT(top_node->type == kDMNodeType_Object_MultiInstance);
    div = &top_node->registered.object_info.inst_vector;

    // Exit if unable to refresh the cache of instances for this object (if necessary)
    err = RefreshInstVector(top_node, true);
    if (err != USP_ERR_OK)
    {
        goto exit;
    }

    // Iterate over the instances array, finding the objects which match, and their instances
    for (i=0; i < div->num_entries; i++)
    {
        oi = &div->vector[i];
        if ((oi->order >= order+1) &&
            (memcmp(oi->nodes, inst->nodes, (order+1)*sizeof(dm_node_t *)) == 0) &&
            (memcmp(oi->instances, inst->instances, order*sizeof(int)) == 0))
        {
            instance = oi->instances[order];

            // Add the instance to the array (if it has not been added already)
            index = INT_VECTOR_Find(iv, instance);
            if (index == INVALID)
            {
                INT_VECTOR_Add(iv, instance);
            }
        }
    }

    err = USP_ERR_OK;

exit:
    inst->nodes[order] = NULL;          // Undo the changes made by this function to the inst array
    return err;
}

/*********************************************************************//**
**
** DM_INST_VECTOR_RefreshBaselineInstances
**
** Called at startup to determine all refreshed object instances, so that
** object creation and deletion after bootup can generate notification events if necessary
** NOTE: This function is called recursively over the whole data model
**
** \param   parent - node to get instances of (if it is a top level multi-instance object)
**
** \return  None - If instances could not be refreshed, then we will try again at the time of generating a USP message that needs them
**
**************************************************************************/
void DM_INST_VECTOR_RefreshBaselineInstances(dm_node_t *parent)
{
    dm_node_t *child;

    // Stop recursing at all top level multi-instance object nodes of the data model tree
    if (parent->type == kDMNodeType_Object_MultiInstance)
    {
        dm_object_info_t *info;                   // Objects

        // Refresh the instances of this object (and all of its children) if required
        info = &parent->registered.object_info;
        if (info->refresh_instances_cb != NULL)
        {
            (void) RefreshInstVector(parent, false);
        }

        USP_ASSERT(parent->order == 1);
        return;
    }

    // Iterate over list of children
    child = (dm_node_t *) parent->child_nodes.head;
    while (child != NULL)
    {
        // Recurse through all objects to find first top level multi-instance object from here
        if (IsObject(child))
        {
            (void) DM_INST_VECTOR_RefreshBaselineInstances(child);
        }

        // Move to next sibling in the data model tree
        child = (dm_node_t *) child->link.next;
    }
}

/*********************************************************************//**
**
** DM_INST_VECTOR_Dump
**
** Prints out the Object Instances array
**
** \param   div - pointer to dm_instances vector structure
**
** \return  None
**
**************************************************************************/
void DM_INST_VECTOR_Dump(dm_instances_vector_t *div)
{
    int i;
    dm_instances_t *inst;
    dm_node_t *node;
    char path[MAX_DM_PATH];

    for (i=0; i < div->num_entries; i++)
    {
        inst = &div->vector[i];
        USP_ASSERT(inst->order >= 1);
        node = inst->nodes[inst->order - 1];
        DM_PRIV_FormInstantiatedPath(node->path, inst, path, sizeof(path));

        USP_DUMP("%s", path);
    }
}

/*********************************************************************//**
**
** DM_INST_VECTOR_GetAllInstancePaths_Unqualified
**
** Returns a string vector containing the paths of all instances to the specified
** unqualified multi-instance object and recursively all child instances
** This function expects the dm_instances_t structure to contain only the node's parents and parent instances
**
** \param   node - pointer to multi-instance object in data model that we want to get the instances of
** \param   inst - pointer to instance structure specifying the object's parents and their instance numbers
** \param   sv - pointer to structure which will be populated with paths to the instances of the object by this function
**               NOTE: The caller must initialise this structure. This function adds to this structure, it does not initialise it.
** \param   combined_role - role to use to check that object instances may be returned.  If set to INTERNAL_ROLE, then full permissions are always returned
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
int DM_INST_VECTOR_GetAllInstancePaths_Unqualified(dm_node_t *node, dm_instances_t *inst, str_vector_t *sv, combined_role_t *combined_role)
{
    int i;
    int order;
    int err;
    dm_instances_t *oi;
    dm_node_t *top_node;
    dm_instances_vector_t *div;

    order = inst->order;          // NOTE: This may be 0 for a top level multi-instance node
    USP_ASSERT(order < MAX_DM_INSTANCE_ORDER);
    inst->nodes[order] = node;

    // Determine which top level multi-instance node's DM instances array to iterate over
    top_node = inst->nodes[0];
    USP_ASSERT(top_node != NULL);
    USP_ASSERT(top_node->type == kDMNodeType_Object_MultiInstance);
    div = &top_node->registered.object_info.inst_vector;

    // Exit if unable to refresh the cache of instances for this object (if necessary)
    err = RefreshInstVector(top_node, true);
    if (err != USP_ERR_OK)
    {
        goto exit;
    }

    // Iterate over the instances array, finding all objects which match, and their instances
    for (i=0; i < div->num_entries; i++)
    {
        oi = &div->vector[i];
        if ((oi->order >= order+1) &&
            (memcmp(oi->nodes, inst->nodes, (order+1)*sizeof(dm_node_t *)) == 0) &&
            (memcmp(oi->instances, inst->instances, order*sizeof(int)) == 0))
        {
            AddObjectInstanceIfPermitted(oi, sv, combined_role);
        }
    }

    err = USP_ERR_OK;

exit:
    // Undo the changes made by this function to the inst array
    inst->nodes[order] = NULL;
    return err;
}

/*********************************************************************//**
**
** DM_INST_VECTOR_GetAllInstancePaths_Qualified
**
** Returns a string vector containing the paths of all instances to the specified
** qualified multi-instance object and recursively all child instances
** This function expects the dm_instances_t structure to contain the object instances to match
**
** \param   inst - pointer to instance structure specifying the object and instance numbers to match
** \param   sv - pointer to structure which will be populated with paths to the instances of the object by this function
**               NOTE: The caller must initialise this structure. This function adds to this structure, it does not initialise it.
** \param   combined_role - role to use to check that object instances may be returned.  If set to INTERNAL_ROLE, then full permissions are always returned
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
int DM_INST_VECTOR_GetAllInstancePaths_Qualified(dm_instances_t *inst, str_vector_t *sv, combined_role_t *combined_role)
{
    int i;
    int order;
    int err;
    dm_instances_t *oi;
    dm_node_t *top_node;
    dm_instances_vector_t *div;

    order = inst->order;
    USP_ASSERT(order > 0);
    USP_ASSERT(order < MAX_DM_INSTANCE_ORDER);

    // Determine which top level multi-instance node's DM instances array to iterate over
    top_node = inst->nodes[0];
    USP_ASSERT(top_node != NULL);
    USP_ASSERT(top_node->type == kDMNodeType_Object_MultiInstance);
    div = &top_node->registered.object_info.inst_vector;

    // Exit if unable to refresh the cache of instances for this object (if necessary)
    err = RefreshInstVector(top_node, true);
    if (err != USP_ERR_OK)
    {
        return err;
    }

    // Iterate over the instances array, finding all objects which match, and their instances
    for (i=0; i < div->num_entries; i++)
    {
        oi = &div->vector[i];
        if ((oi->order >= order) &&
            (memcmp(oi->nodes, inst->nodes, order*sizeof(dm_node_t *)) == 0) &&
            (memcmp(oi->instances, inst->instances, order*sizeof(int)) == 0))
        {
            AddObjectInstanceIfPermitted(oi, sv, combined_role);
        }
    }

    return USP_ERR_OK;
}

/*********************************************************************//**
**
** DM_INST_VECTOR_RefreshInstance
**
** Adds the specified object instance into the instance vector
** NOTE: This function may only be called by the vendor within the context of the get_instances_cb call
**       It must contain only instances of the object specified in the get_instances_cb (and that object's children)
**
** \param   path - data model path of the multi-instance object to add
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
int DM_INST_VECTOR_RefreshInstance(char *path)
{
    dm_node_t *node;
    dm_node_t *top_node;
    dm_instances_t inst;
    bool is_qualified_instance;
    bool exists;

    // Exit if unable to find node representing this object
    node = DM_PRIV_GetNodeFromPath(path, &inst, &is_qualified_instance);
    if (node == NULL)
    {
        USP_ERR_SetMessage("%s: Path (%s) does not exist in the schema", __FUNCTION__, path);
        return USP_ERR_OBJECT_DOES_NOT_EXIST;
    }

    // Exit if the object the vendor signalled was not a multi-instance object
    if (node->type != kDMNodeType_Object_MultiInstance)
    {
        USP_ERR_SetMessage("%s: Path (%s) is not a multi-instance object.", __FUNCTION__, path);
        return USP_ERR_OBJECT_NOT_CREATABLE;
    }

    // Exit if this object is not a fully qualified instance
    if (is_qualified_instance == false)
    {
        USP_ERR_SetMessage("%s: Path (%s) should contain instance number of object that was added", __FUNCTION__, path);
        return USP_ERR_INVALID_ARGUMENTS;
    }

    // Exit if the top-level multi-instance object does not match the one we requested instances for
    // NOTE: We just emit a warning in this case, rather than returning an error
    // This allows clients to potentially be 'dumb' in refreshing instances - we filter the instances to add here
    top_node = inst.nodes[0];
    if (top_node != refresh_instances_top_node)
    {
        USP_LOG_Warning("%s: Ignoring USP_RefreshInstance(%s) as it was not for path %s (or its descendants)", __FUNCTION__, path, top_node->path);
        return USP_ERR_OK;
    }

    // Exit if instance already exists - nothing to do
    exists = IsExistInInstVector(&inst, &refreshed_instances_vector);
    if (exists)
    {
        return USP_ERR_OK;
    }

    // Exit if the parent object instances in the path do not exist
    if (inst.order > 1)
    {
        inst.order--;           // Temporarily remove the instance number of the object that was added,
                                // so that structure indicates only parent instance numbers
        exists = IsExistInInstVector(&inst, &refreshed_instances_vector);
        if (exists == false)
        {
            USP_ERR_SetMessage("%s: Parent objects in path (%s) do not exist", __FUNCTION__, path);
            return USP_ERR_OBJECT_DOES_NOT_EXIST;
        }
        inst.order++;           // Restore the structure, so that it indicates the instance number of the object that was added
    }

    // Add this to the refreshed instances vector
    AddToInstVector(&inst, &refreshed_instances_vector);

    return USP_ERR_OK;
}

/*********************************************************************//**
**
** DM_INST_VECTOR_RefreshTopLevelObjectInstances
**
** Refreshes the instances for the specified top level object and all children
** NOTE: This function may be called recursively, since it is called from the path resolver and it may call
**       the path resolver itself in order to resolve object lifetime event subscriptions
**
** \param   node - pointer to node of top level object to refresh
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
int DM_INST_VECTOR_RefreshTopLevelObjectInstances(dm_node_t *node)
{
    return RefreshInstVector(node, true);
}

/*********************************************************************//**
**
** AddObjectInstanceIfPermitted
**
** Adds the specified object instance, to the string vector, if the role permits its instance numbers to be read
**
** \param   inst - pointer to instance structure specifying the object and its instance numbers
** \param   sv - pointer to structure which will be populated with paths to the instances of the object by this function
**               NOTE: The caller must initialise this structure. This function adds to this structure, it does not initialise it.
** \param   combined_role - role to use to check that object instances may be returned.  If set to INTERNAL_ROLE, then full permissions are always returned
**
** \return  None
**
**************************************************************************/
void AddObjectInstanceIfPermitted(dm_instances_t *inst, str_vector_t *sv, combined_role_t *combined_role)
{
    dm_node_t *node;
    unsigned short permission_bitmask;
    char path[MAX_DM_PATH];
    int err;

    // Exit if the current role does not have permission to return this object instance in the string vector
    node = inst->nodes[inst->order-1];
    permission_bitmask = DM_PRIV_GetPermissions(node, combined_role);
    if ((permission_bitmask & PERMIT_GET_INST)==0)
    {
        return;
    }

    // Convert the dm_instances_t structure into a path
    err = DM_PRIV_FormInstantiatedPath(node->path, inst, path, sizeof(path));
    if (err != USP_ERR_OK)
    {
        return;
    }

    // Add the path to the string vector
    STR_VECTOR_Add(sv, path);
}

/*********************************************************************//**
**
** RefreshInstVector
**
** Called before querying the instances vector, to ensure that it is up to date
** If a refresh_instances_cb is registered, then the instances vector acts as a cache, with ageing of its content
** In this case, this function may call the refresh_instances_cb to obtain the instance numbers if the current ones in the cache have expired
**
** \param   top_node - pointer to top-level multi-instance object
** \param   notify_subscriptions - set if subscriptions should be notified of instances added/deleted.
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
int RefreshInstVector(dm_node_t *top_node, bool notify_subscriptions)
{
    dm_instances_vector_t *old_instances;
    dm_instances_vector_t *new_instances;
    dm_instances_vector_t deleted_instances;
    dm_instances_t *inst;
    dm_node_t *node;
    char path[MAX_DM_PATH];
    dm_object_info_t *info;
    time_t cur_time;
    int expiry_period;
    int err;
    int len;
    int i;

    // Exit if this function is being called re-entrantly
    // This may be the case, as the refresh instances vendor callback ends up calling DM_INST_VECTOR_RefreshInstance()
    if (refresh_instances_top_node != NULL)
    {
        return USP_ERR_OK;
    }

    // Exit if this top-level multi-instance node is not maintained by the get instances callback
    // NOTE: Objects which don't have a refresh instances callback use USP_DM_InformInstance(), USP_SIGNAL_ObjectAdded() and USP_SIGNAL_ObjectDeleted() to maintain the instance vector
    info = &top_node->registered.object_info;
    if (info->refresh_instances_cb == NULL)
    {
        return USP_ERR_OK;
    }

    // Exit if it's not yet time to refresh the instance vector
    cur_time = time(NULL);
    if (cur_time <= info->refresh_instances_expiry_time)
    {
        return USP_ERR_OK;
    }

    // If the code gets here, then the instances must be refreshed

    // Truncate the schema path to form a partial path to the top-level multi-instance object
    #define INSTANCE_SEPARATOR "{i}"
    len = strlen(top_node->path) - (sizeof(INSTANCE_SEPARATOR)-1);
    USP_ASSERT(strcmp(&top_node->path[len], INSTANCE_SEPARATOR)==0);
    memcpy(path, top_node->path, len);
    path[len] = '\0';


    // Exit if unable to get the refreshed instances into the refreshed_instances_vector
    refresh_instances_top_node = top_node;      // Indicate to DM_INST_VECTOR_RefreshInstance() the top level node which is meant to be being refreshed
    DM_INST_VECTOR_Init(&refreshed_instances_vector);
    err = info->refresh_instances_cb(info->group_id, path, &expiry_period);
    if (err != USP_ERR_OK)
    {
        USP_ERR_ReplaceEmptyMessage("%s: Refresh Instances callback for %s failed", __FUNCTION__, top_node->path);
        DM_INST_VECTOR_Destroy(&refreshed_instances_vector);
        refresh_instances_top_node = NULL;
        return USP_ERR_INTERNAL_ERROR;
    }

    // Update the expiry time
    cur_time = time(NULL);
    info->refresh_instances_expiry_time = cur_time + expiry_period;

    // Skip determining add added/deleted instances, if we don't need to notify subscriptions because
    // we're getting the baseline set of object instances at bootup
    if (notify_subscriptions == false)
    {
        goto exit;
    }

    // Determine all instances which have been added, by finding all instances in the new, which are not in the old instance vector
    old_instances = &info->inst_vector;
    new_instances = &refreshed_instances_vector;
    for (i=0; i < new_instances->num_entries; i++)
    {
        inst = &new_instances->vector[i];
        if (IsExistInInstVector(inst, old_instances)==false)
        {
            node = inst->nodes[inst->order-1];
            err = DM_PRIV_FormInstantiatedPath(node->path, inst, path, sizeof(path));
            if (err == USP_ERR_OK)
            {
                DEVICE_SUBSCRIPTION_NotifyObjectLifeEvent(path, kSubNotifyType_ObjectCreation);
            }
        }
    }

    // Create a vector of all instances which have been deleted, by finding all instances in the old, which are not in the new instance vector
    // NOTE: For why we create a vector, rather than calling DEVICE_SUBSCRIPTION_NotifyObjectLifeEvent() directly, see below
    DM_INST_VECTOR_Init(&deleted_instances);
    for (i=0; i < old_instances->num_entries; i++)
    {
        inst = &old_instances->vector[i];
        if (IsExistInInstVector(inst, new_instances)==false)
        {
            AddToInstVector(inst, &deleted_instances);
        }
    }

    // NOTE: We need to call DEVICE_SUBSCRIPTION_ResolveObjectDeletionPaths() before DEVICE_SUBSCRIPTION_NotifyObjectLifeEvent()
    //       But as that is an unnecessary (and costly) operation if no instances are deleted, we only do it if there were any instances that were deleted
    if (deleted_instances.num_entries > 0)
    {
        // NOTE: DEVICE_SUBSCRIPTION_ResolveObjectDeletionPaths() must be called before any objects are actually deleted from the data model
        // NOTE: DEVICE_SUBSCRIPTION_ResolveObjectDeletionPaths() calls RefreshInstVector() recursively, so refresh_instances_top_node must still be set here
        DEVICE_SUBSCRIPTION_ResolveObjectDeletionPaths();

        for (i=0; i<deleted_instances.num_entries; i++)
        {
            inst = &deleted_instances.vector[i];
            node = inst->nodes[inst->order-1];
            err = DM_PRIV_FormInstantiatedPath(node->path, inst, path, sizeof(path));
            if (err == USP_ERR_OK)
            {
                DEVICE_SUBSCRIPTION_NotifyObjectLifeEvent(path, kSubNotifyType_ObjectDeletion);
            }
        }
    }
    DM_INST_VECTOR_Destroy(&deleted_instances);

exit:
    // Replace the old instances with the new instances, deleting the old instances first
    DM_INST_VECTOR_Destroy(&info->inst_vector);
    memcpy(&info->inst_vector, &refreshed_instances_vector, sizeof(dm_instances_vector_t));
    refresh_instances_top_node = NULL;

    return USP_ERR_OK;
}

/*********************************************************************//**
**
** IsExistInInstVector
**
** Determines whether the specified object instance exists in the specified instances vector
**
** \param   match - pointer to instances structure describing the instances to match against
** \param   div - pointer to dm_instances vector structure to search in
**
** \return  USP_ERR_OK if successful
**
**************************************************************************/
bool IsExistInInstVector(dm_instances_t *match, dm_instances_vector_t *div)
{
    int i;
    dm_instances_t *inst;

    // Iterate over the array of object instances
    for (i=0; i < div->num_entries; i++)
    {
        inst = &div->vector[i];
        if (inst->order >= match->order)
        {
            if ( (memcmp(inst->nodes, match->nodes, (match->order)*sizeof(dm_node_t *)) == 0) &&
                 (memcmp(inst->instances, match->instances, (match->order)*sizeof(int)) == 0) )
            {
                // All specified object instances match
                return true;
            }
        }
    }

    // If the code gets here, then no instances matched
    return false;
}

/*********************************************************************//**
**
** AddToInstVector
**
** Adds the specified object instance into the specified instances vector
**
** \param   inst - pointer to instances structure describing the object instance
** \param   div - pointer to dm_instances vector structure to add to
**
** \return  None
**
**************************************************************************/
void AddToInstVector(dm_instances_t *inst, dm_instances_vector_t *div)
{
    int size;

    // Increase the size of the dm_instances_vector array
    size = (div->num_entries+1) * sizeof(dm_instances_t);
    div->vector = USP_REALLOC(div->vector, size);

    // And store this object instance
    memcpy(&div->vector[div->num_entries], inst, sizeof(dm_instances_t));
    div->num_entries++;
}

